/* Copyright 1999 by Donald J. Bindner */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "convert.h"

#define NUMBER(array) (sizeof(array)/sizeof(array[0]))
#define TRUE 1
#define FALSE 0

typedef signed char BYTE;

/* Unit prefixes.  Each prefix is a single letter.  They are in
 * order from highest magnitude to lowest.
 */
static char *prefixes = "EPTGMKkHhDdcmnpfa";

/* The magnitudes associated with each prefix */
static double pre_factor[] = { 1e18, 1e15, 1e12, 1e9, 1e6, 1e3, 1e3, 1e2, 1e2,
                               1e1, 1e-1, 1e-2, 1e-3, 1e-6, 1e-9, 1e-12, 1e-15,
                               1e-18 };

/* To save memory, common descriptions of units are only defined
 * one time.  From then on we use pointers to the discriptions.
 */
static char s_area[] = "Area";
static char s_angle[] = "Plane angle";
static char s_pressure[] = "Pressure";
static char s_length[] = "Length";
static char s_volume[] = "Volume";
static char s_force[] = "Force";
static char s_energy[] = "Energy";
static char s_mass[] = "Mass";
static char s_speed[] = "Speed";
static char s_charge[] = "Electric charge";
static char s_time[] = "Time";
static char s_temp[] = "Temperature";
static char s_power[] = "Power";
static char s_lum[] = "Luminance";
static char s_adose[] = "Absorbed dose";
static char s_mflux[] = "Magnetic flux";

/* The names of the fundamental units */
static char *fun_name[] = { "A", "cd", "kg", "m", "s", "K" };

/* Again to save memory, each type of unit is only defined one time.
 * each element of the array records how many times each fundament unit
 * occurs in its definition:
 *                             A cd kg  m  s K
 */
static BYTE area[] =         { 0, 0, 0, 2, 0, 0};
static BYTE current[] =      { 1, 0, 0, 0, 0, 0};
static BYTE none[] =         { 0, 0, 0, 0, 0, 0};
static BYTE pressure[] =     { 0, 0, 1,-1,-2, 0};
static BYTE length[] =       { 0, 0, 0, 1, 0, 0};
static BYTE volume[] =       { 0, 0, 0, 3, 0, 0};
static BYTE freq[] =         { 0, 0, 0, 0,-1, 0};
static BYTE force[] =        { 0, 0, 1, 1,-2, 0};
static BYTE energy[] =       { 0, 0, 1, 2,-2, 0};
static BYTE mass[] =         { 0, 0, 1, 0, 0, 0};
static BYTE velocity[] =     { 0, 0, 0, 1,-1, 0};
static BYTE charge[] =       { 1, 0, 0, 0, 1, 0};
static BYTE time[] =         { 0, 0, 0, 0, 1, 0};
static BYTE temp[] =         { 0, 0, 0, 0, 0, 1};
static BYTE power[] =        { 0, 0, 1, 2,-3, 0};
static BYTE candela[] =      { 0, 1, 0, 0, 0, 0};
static BYTE capac[] =        { 2, 0,-1,-2, 4, 0};
static BYTE induct[] =       {-2, 0, 1, 2,-2, 0};
static BYTE luminance[] =    { 0, 1, 0,-2, 0, 0};
static BYTE accel[] =        { 0, 0, 0, 1,-2, 0};
static BYTE ohm[] =          {-2, 0, 1, 2,-3, 0};
static BYTE rad[] =          { 1, 0,-1, 0, 1, 0};
static BYTE adose[] =        { 0, 0, 0, 2,-2, 0};
static BYTE volt[] =         {-1, 0, 1, 2,-3, 0};
static BYTE dose_eq[] =      { 0, 0, 0, 2,-2, 0};
static BYTE mflux[] =        {-1, 0, 1, 0,-2, 0};
static BYTE wflux[] =        {-1, 0, 1, 2,-2, 0};
static BYTE phot[] =         { 0, 1, 0,-2, 0, 0};
static BYTE visc[] =         { 0, 0, 1,-1,-1, 0};
static BYTE kvisc[] =        { 0, 0, 0, 2,-1, 0};
static BYTE mho[] =          { 2, 0,-1,-2, 3, 0};

/* The master list.  This contains the fields for:
 *  abbreviated unit name
 *  full unit name
 *  description
 *  type of unit ( as an array of indices of fundamental units )
 *  factor ( to multiply to convert to fundamental units )
 */
UNIT unit[] = {
    { "1", "Dimensionless unit","", none, 1.0 },
    { "A", "Ampere", "Electric current", current, 1.0 },
    { "A", "Angstrom", s_length, length, 1e-10 },
    { "Bq", "Becquerel", "Activity", freq, 1.0 },
    { "Btu", "International Table Btu", s_energy, energy, 1055.05585262 },
    { "C", "Coulomb", s_charge, charge, 1.0 },
    { "Ci", "Curie", "Activity", freq, 3.7e10 },
    { "F", "Farad", "Capacitance", capac, 1.0 },
    { "Fdy", "Faraday", s_charge, charge, 96487.0 },
    { "Gy", "Gray", s_adose, adose, 1.0 },
    { "H", "Henry", "Inductance", induct, 1.0 },
    { "Hz", "Hertz", "Frequency", freq, 1.0 },
    { "J", "Joule", s_energy, energy, 1.0 },
    { "N", "Newton", s_force, force, 1.0 },
    { "P", "Poise", "Dynamic viscosity", visc, .1 },
    { "Pa", "Pascal", s_pressure, pressure, 1.0 },
    { "R", "Roentgen", "Radiation exposure", rad, .000258 },
    { "S", "Siemens", "Electric conductance", mho, 1.0 },
    { "St", "Stokes", "Kinematic viscosity", kvisc, .0001 },
    { "Sv", "Sievert", "Dose equivalant", dose_eq, 1.0 },
    { "T", "Tesla", s_mflux, mflux, 1.0 },
    { "V", "Volt", "Electric potential", volt, 1.0 },
    { "W", "Watt", s_power, power, 1.0 },
    { "Wb", "Weber", s_mflux, wflux, 1.0 },
    { "a", "Are", s_area, area, 100.0 },
    { "acre", "Acre", s_area, area, 4046.87260987 },
    { "arcmin", "Minute of arc", s_angle, none, 4.629629629629629e-5 },
    { "arcs", "Second of arc", s_angle, none, 7.71604938272e-7 },
    { "atm", "Atmosphere", s_pressure, pressure, 101325.0 },
    { "au", "Astronomical unit", s_length, length, 149597900000.0 },
    { "b", "Barn", s_area, area, 1e-28 },
    { "bar", "Bar", s_pressure, pressure, 1e5 },
    { "bbl", "Barrel, oil", s_volume, volume, .158987294928 },
    { "bu", "Bushel", s_volume, volume, .03523907 },
    { "c", "Speed of light", "Velocity", velocity, 299792458.0 },
    { "cal", "International Table calorie", s_energy, energy, 4.1868 },
    { "cd", "Candela", "Luminous intensity", candela, 1.0 },
    { "chain", "Chain", s_length, length, 20.1168402337 },
    { "ct", "Carat", s_mass, mass, .0002 },
    { "cu", "US cup", s_volume, volume, 2.365882365e-4 },
    { "d", "Day", s_time, time, 86400.0 },
    { "dyn", "Dyne", s_force, force, .00001 },
    { "eV", "Electron volt", s_energy, energy, 1.60219e-19 },
    { "erg", "Erg", s_energy, energy, .0000001 },
    { "fath", "Fathom", s_length, length, 1.82880365761 },
    { "fbm", "Board foot", s_volume, volume, .002359737216 },
    { "fc", "Footcandle", s_lum, none, .856564774909 },
    { "fermi", "Fermi", s_length, length, 1e-15 },
    { "flam", "Footlambert", s_lum, luminance, 3.42625909964 },
    { "ft", "International foot", s_length, length, .3048 },
    { "ftUS", "Survey foot", s_length, length, .304800609601 },
    { "g", "Gram", s_mass, mass, .001 },
    { "ga", "Standard freefall", "Acceleration", accel, 9.80665 },
    { "gal", "US gallon", s_volume, volume, .003785411784 },
    { "galC", "Canadian gallon", s_volume, volume, .00454609 },
    { "galUK", "UK gallon", s_volume, volume, .004546092 },
    { "gf", "Gram-force", s_force, force, .00980665 },
    { "grad", "Grade", s_angle, none, .0025 },
    { "grain", "Grain", s_mass, mass, .00006479891 },
    { "h", "Hour", s_time, time, 3600.0 },
    { "hp", "Horsepower", s_power, power, 745.699871582 },
    { "in", "Inch", s_length, length, .0254 },
    { "inH2O", "Inches of water", s_pressure, pressure, 248.84 },
    { "inHg", "Inches of mercury", s_pressure, pressure, 3386.38815789 },
    { "kip", "Kilopound-force", s_force, force, 4448.22161526 },
    { "knot", "Knot", s_speed, velocity, .514444444444 },
    { "kph", "Kilometer per hour", s_speed, velocity, .277777777778 },
    { "l", "Liter", s_volume, volume, .001 },
    { "lam", "Lambert", s_lum, luminance, 3183.09886184 },
    { "lb", "Avoirdupois pound", s_mass, mass, .45359237 },
    { "lbf", "Pound-force", s_force, force, 4.44822161526 },
    { "lbt", "Troy lb", s_mass, mass, .3732417 },
    { "lm", "Lumen", "Luminance flux", candela, 7.95774715459e-2 },
    { "lx", "Lux", "Illuminance", luminance, 7.95774715459e-2 },
    { "lyr", "Light year", s_length, length, 9.46052840488e15 },
    { "m", "Meter", s_length, length, 1.0 },
    { "mho", "Mho", "Electric conductance", mho, 1.0 },
    { "mi", "International mile", s_length, length, 1609.344 },
    { "miUS", "US statute mile", s_length, length, 1609.34721869 },
    { "mil", "Mil", s_length, length, .0000254 },
    { "min", "Minute", s_time, time, 60.0 },
    { "mmHg", "Millimeter of mercury", s_pressure, pressure, 133.322368421 },
    { "mph", "Miles per hour", s_speed, velocity, .44704 },
    { "nmi", "Nautical mile", s_length, length, 1852.0 },
    { "ohm", "Ohm", "Electric resistance", ohm, 1.0 },
    { "oz", "Ounce", s_mass, mass, .028349523125 },
    { "ozUK", "UK fluid oz", s_volume, volume, .000028413075 },
    { "ozfl", "US fluid oz", s_volume, volume, 2.95735295625e-5 },
    { "ozt", "Troy oz", s_mass, mass, .031103475 },
    { "pc", "Parsec", s_length, length, 3.08567818585e16 },
    { "pdl", "Poundal", s_force, force, .138254954376 },
    { "ph", "Phot", s_lum, phot, 795.774715459 },
    { "pk", "Peck", s_volume, volume, .0088097675 },
    { "psi", "Pounds per square inch", s_pressure, pressure, 6894.75729317 },
    { "pt", "Pint", s_volume, volume, .000473176473 },
    { "qt", "Quart", s_volume, volume, .000946352946 },
    { "r", "Radian", s_angle, none, .159154943092 },
    { "rad", "Rad", s_adose, adose, .01 },
    { "rd", "Rod", s_length, length, 5.02921005842 },
    { "rem", "Rem", "Dose equivalent", adose, .01 },
    { "s", "Second", s_time, time, 1.0 },
    { "sb", "Stilb", s_lum, luminance, 10000.0 },
    { "slug", "Slug", s_mass, mass, 14.5939029372 },
    { "sr", "Steradian", "Solid angle", none, 7.95774715459e-2 },
    { "st", "Stere", s_volume, volume, 1.0 },
    { "t", "Metric ton", s_mass, mass, 1000.0 },
    { "tbsp", "Tablespoon", s_volume, volume, 1.47867647813e-5 },
    { "therm", "EEC therm", s_energy, energy, 105506000.0 },
    { "ton", "Short ton", s_mass, mass, 907.18474 },
    { "tonUK", "Long ton", s_mass, mass, 1016.0469088 },
    { "torr", "Torr", s_pressure, pressure, 133.322368421 },
    { "tsp", "Teaspoon", s_volume, volume, 4.92892159375e-6 },
    { "u", "Unified atomic mass", s_mass, mass, 1.66057e-27 },
    { "yd", "International yard", s_length, length, .9144 },
    { "yr", "Year", s_time, time, 31556925.9747 },
    { "", "Micron", s_length, length, .000001 },
    { "", "Degree", "Angle", none, 2.7777777777777777777e-3 },
    { "C", "Degree Celsius", s_temp, temp, 1.0 },
    { "F", "Degree Fahrenheit", s_temp, temp, .555555555555555555 },
    { "K", "Degree Kelvin", s_temp, temp, 1.0 },
    { "R", "Degree Rankine", s_temp, temp, .555555555555555555 }};

/* Pulls the first word out of a string.  It assumes that
 * the string points to a valid word, and not
 * an asterisk or something.
 */
static char *first_word( char *ptr, char *subword, int *num ) {
    int i = 0;
    *num = 1;

    while( i < 19 ) {
        if( *ptr == 0 ) break;

        subword[i++] = *(ptr++);

        if( *ptr == '^' ) {
            ptr++;
            *num = atoi( ptr );
            while( isdigit( *ptr )) ptr++;
            break;
        }

        if(( *ptr == '*' ) || ( *ptr == '/' )) break;
    }

    subword[i] = 0;
    return ptr;
}

/* Finds the index in the prefix string for the character received,
 * or -1 if that character is not in the string.  The result could
 * then be used to look up the multiplication factor for that prefix.
 */
static int find_prefix( char ch ) {
    int i;

    for( i = 0; i < 18; i++ ) {
        if( ch == prefixes[i] ) return i;
    }
    return -1;
}

/* Does a quick binary search of the master list to find the unit
 * referred to by the string ident.  It returns TRUE if found and
 * the type, and conversion factor.
 */
static UNIT *lookup( char *ident ) {
    int low = 0, high = NUMBER( unit ), mid, cmp;

    while( low <= high ) {
        mid = ( low + high ) / 2;
        if( !( cmp = strcmp( ident, unit[mid].name ))) {
            return (unit + mid);
        }
        if( cmp < 0 ) {
            high = mid-1;
        } else {
            low = mid+1;
        }
    }
    return NULL;
}

int temperature_convert( char *in_units, char *out_units, double *conversion ) {
    if(( strlen( in_units ) != 2 ) || ( strlen( out_units ) != 2 )) return FALSE;
    if( strstr( "C F K R", in_units ) == NULL ) return FALSE;
    if( strstr( "C F K R", out_units ) == NULL ) return FALSE;

    switch( in_units[1] ) {
        case 'K': break;
        case 'C': *conversion += 273.15;
                  break;
        case 'R': *conversion *= 5.0/9.0;
                  break;
        case 'F': *conversion = ((*conversion) + 459.67) * 5.0 / 9.0;
                  break;
    }

    switch( out_units[1] ) {
        case 'K': break;
        case 'C': *conversion -= 273.15;
                  break;
        case 'R': *conversion *= 9.0/5.0;
                  break;
        case 'F': *conversion = (*conversion) * 9.0 / 5.0 - 459.67;
                  break;
    }
    return TRUE;
}


/* Uses the other routines to perform the unit conversion.  Returns
 * TRUE if successful.  The double conversion will contain the
 * neccessary factor to convert from the 1st units to the 2nd.
 */
int convert( char *in_units, char *out_units, double *conversion ) {
    char ident[20];
    int exponent, prefix, i, j, sign = 1;
    char *ptr = in_units;
    signed char temp_type[6];
    double t_conv, p_factor;
    UNIT *u_ptr;

    if( temperature_convert( in_units, out_units, conversion )) return TRUE;

    for( i = 0; i < 6; i++ ) {
        temp_type[i] = 0;
    }

    /* Loop first time to convert to standard units and
     * then again to convert to the second set of units
     */
    for( j = 0; j < 2; j++ ) {
        t_conv = 1.0;

        /* choose the string to convert */
        if( !j ) ptr = in_units; else ptr = out_units;

        /* parse the numerator */
        while( 1 ) {
            if( *ptr == 0 ) break;

            /* pull the first word of the string and increment the pointer */
            ptr = first_word( ptr, ident, &exponent );

            p_factor = 1.0;
            /* look up the possible unit indentifier */
            if(( u_ptr = lookup( ident )) == NULL ) {
                /* since we didn't find it, there may be a prefix */
                if(( prefix = find_prefix( *ident )) == -1 ) {
                    return FALSE;
                } else {
                    /* look up the word with prefix removed */
                    if(( u_ptr = lookup( ident+1 )) == NULL ) {
                        return FALSE;
                    }
                    p_factor = pre_factor[ prefix ];
                }
            }
            /* temperary conversion factor -- remember the conversion
             * for the unit we just looked up.
             */
            t_conv *= pow( p_factor * u_ptr->factor, (double)exponent );

            /* remember the unit type so we can check that the conversion
             * is consistent.
             */
            for( i = 0; i < 6; i++ ) {
                temp_type[i] += (signed char)( sign * u_ptr->type[i] * exponent  );
            }

            if( *ptr == '*' ) {
                ptr++;
            } else if( *ptr == '/' ) {
                ptr++;
                break;
            }
        }

        /* parse the denominator */
        while( 1 ) {
            if( *ptr == 0 ) break;

            /* pull the first word of the string and increment the pointer */
            ptr = first_word( ptr, ident, &exponent );

            p_factor = 1.0;
            /* look up the possible unit indentifier */
            if(( u_ptr = lookup( ident )) == NULL ) {
                /* since we didn't find it, there may be a prefix */
                if(( prefix = find_prefix( *ident )) == -1 ) {
                    return FALSE;
                } else {
                    /* look up the word with prefix removed */
                    if(( u_ptr = lookup( ident+1 )) == NULL ) {
                        return FALSE;
                    }
                    p_factor = pre_factor[ prefix ];
                }
            }
            /* temperary conversion factor -- remember the conversion
             * for the unit we just looked up.  We divide now since
             * this is the denominator.
             */
            t_conv /= pow( p_factor * u_ptr->factor, (double)exponent );

            /* remember the unit type so we can check that the conversion
             * is consistent.  We subtract this time since it is in the
             * denominator.
             */
            for( i = 0; i < 6; i++ ) {
                temp_type[i] -= (signed char)( sign * u_ptr->type[i] * exponent  );
            }

            if( *ptr == '*' ) {
                ptr++;
            }
        }

        sign = -sign;

        if( !j ) {
            /* convert from 1st units to standard units */
            *conversion *= t_conv;
        } else {
            /* convert from standard units to 2nd units */
            *conversion /= t_conv;
        }
    }

    /* if the units don't match here exactly then this whole
     * calculation was bogus so return FALSE
     */
    for( i = 0; i < 6; i++ ) {
        if( temp_type[i] ) return FALSE;
    }
    return TRUE;
}

char *unit_string( int index ) {
    int flag1=0, flag2=0, i;
    static char *t = "0";
    static char out[20];

    *out = 0;

    for( i = 0; i < 6; i++ ) {
        if( unit[index].type[i] > 0 ) {
            if( flag1 ) strcat( out, "*" ); else flag1++;

            strcat( out, fun_name[i] );

            if( unit[index].type[i] > 1 ) {
                strcat( out, "^" );
                *t = (char)( unit[index].type[i] + '0' );
                strcat( out, t );
            }

        } else if( unit[index].type[i] < 0 ) {
            if( !flag2 ) flag2++;
        }
    }

    if( !flag2 ) return out;

    if( !flag1 ) strcat( out, "1" );

    strcat( out, "/" );

    flag2 = 0;
    for( i = 0; i < 6; i++ ) {
        if( unit[index].type[i] < 0 ) {
            if( flag2 ) strcat( out, "*" ); else flag2++;

            strcat( out, fun_name[i] );

            if( unit[index].type[i] < -1 ) {
                strcat( out, "^" );
                *t = (char)( -unit[index].type[i] + '0');
                strcat( out, t );
            }
        }
    }
    return out;
}

